/* event_neonkick_join.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

struct neonkick_event_join_cache {
    struct ClientSocket *client;
    struct ChanUser *chanuser;
};

static USERAUTH_CALLBACK(neonkick_event_join_nick_lookup);
static void neonkick_event_join_async1(struct ClientSocket *client, struct ChanUser *chanuser);

static void neonkick_event_join(struct ChanUser *chanuser) {
    struct UserNode *user = chanuser->user;
    struct ClientSocket *client = getChannelBot(chanuser->chan, BOTID);
    if(!client) return; //we can't "see" this event
    if(chanuser->user == client->user) {
        requestOp(client->user, chanuser->chan);
        module_neonbackup_recover_chan(chanuser->chan);
        return;
    }
    if(chanuser->user->flags & USERFLAG_ISBOT) return;
    loadChannelSettings(chanuser->chan);
    if(!(chanuser->chan->flags & CHANFLAG_CHAN_REGISTERED)) return;
    if(!(user->flags & USERFLAG_ISAUTHED)) {
        struct neonkick_event_join_cache *cache = malloc(sizeof(*cache));
        if (!cache) {
            perror("malloc() failed");
            return;
        }
        cache->client = client;
        cache->chanuser = chanuser;
        get_userauth(user, module_id, neonkick_event_join_nick_lookup, cache);
    } else
        neonkick_event_join_async1(client, chanuser);
}

static USERAUTH_CALLBACK(neonkick_event_join_nick_lookup) {
    struct neonkick_event_join_cache *cache = data;
    if(user)
        neonkick_event_join_async1(cache->client, cache->chanuser);
    free(cache);
}

static void neonkick_event_join_async1(struct ClientSocket *client, struct ChanUser *chanuser) {
    struct ChanNode *chan = chanuser->chan;
    struct UserNode *user = chanuser->user;
    MYSQL_RES *res;
    MYSQL_ROW chanuserrow;
    
    if((user->flags & USERFLAG_ISAUTHED)) {
        printf_mysql_query("SELECT `chanuser_access`, `chanuser_flags`, `chanuser_infoline`, `chanuser_seen`, `chanuser_id` FROM `chanusers` LEFT JOIN `users` ON `chanuser_uid` = `user_id` WHERE `chanuser_cid` = '%d' AND `user_user` = '%s'", chan->channel_id, escape_string(user->auth));
        res = mysql_use();
        chanuserrow = mysql_fetch_row(res);
    } else
        chanuserrow = NULL;
    int userflags = (chanuserrow ? atoi(chanuserrow[1]) : 0);
    int uaccess = ((chanuserrow && !(userflags & DB_CHANUSER_SUSPENDED)) ? atoi(chanuserrow[0]) : 0);
    if(uaccess > 0)
        return; //no kick :D
    char *tmp;
    int global_kick_count = ((tmp = getSetting(NULL, chan, "kicks")) ? atoi(tmp) : 0);
    int kick_count = ((tmp = getSetting(user, chan, "kicks")) ? atoi(tmp) : 0);
    global_kick_count++;
    kick_count++;
    char buf[MAXLEN];
    char reason[MAXLEN];
    sprintf(buf, "%d", global_kick_count);
    setSetting(NULL, chan, "kicks", buf);
    sprintf(buf, "%d", kick_count);
    setSetting(user, chan, "kicks", buf);
    char *banmask = generate_banmask(chanuser->user, buf);
    sprintf(reason, "Kick #%d (#%d for you)", global_kick_count, kick_count);
    putsock(client, "KICK %s %s :%s", chanuser->chan->name, chanuser->user->nick, reason);
    putsock(client, "MODE %s +b %s", chanuser->chan->name, banmask);
    { // add timed ban
        printf_mysql_query("INSERT INTO `bans` (`ban_channel`, `ban_mask`, `ban_triggered`, `ban_timeout`, `ban_owner`, `ban_reason`) VALUES ('%d', '%s', UNIX_TIMESTAMP(), '%lu', '%d', '%s')", chanuser->chan->channel_id, escape_string(banmask), (unsigned long) (time(0) + 10), 0, escape_string(reason));
        int banid = (int) mysql_insert_id(get_mysql_conn());
        char nameBuf[MAXLEN];
        char banidBuf[20];
        sprintf(nameBuf, "ban_%d", banid);
        sprintf(banidBuf, "%d", banid);
        timeq_add_name(nameBuf, 10, module_id, channel_ban_timeout, strdup(banidBuf));
    }
}
