/* cmd_neonserv_noregister.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

#include "cmd_neonserv.h"

/*
* argv[0]   *auth/#channel
* argv[1]   duration
* argv[2-*] reason
*/

static AUTHLOOKUP_CALLBACK(neonserv_cmd_noregister_auth_lookup);
static USERAUTH_CALLBACK(neonserv_cmd_noregister_nick_lookup);
static void neonserv_cmd_noregister_async1(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct ChanNode *chan, struct Event *event,char *auth, int duration, char *reason);
static void neonserv_cmd_noregister_list(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct ChanNode *chan, struct Event *event, char **argv, int argc);

struct neonserv_cmd_noregister_cache {
    struct ClientSocket *client, *textclient;
    struct UserNode *user;
    struct ChanNode *chan;
    struct Event *event;
    int duration;
    char *reason, *nick;
};

CMD_BIND(neonserv_cmd_noregister) {
    if(argc < 3) {
        neonserv_cmd_noregister_list(client, textclient, user, chan, event, argv, argc);
        return;
    }
    int duration = strToTime(user, argv[1]);
    if(duration == 0 && strcmp(argv[1], "0")) {
        reply(textclient, user, "NS_NOREGISTER_INVALID_DURATION", argv[1]);
        return;
    }
    char *reason = merge_argv(argv, 2, argc);
    MYSQL_RES *res;
    MYSQL_ROW row;
    if(argv[0][0] == '#') {
        printf_mysql_query("SELECT `botid` FROM `bot_channels` LEFT JOIN `channels` ON `bot_channels`.`chanid` = `channels`.`channel_id` WHERE `channel_name` = '%s'", escape_string(argv[0]));
        res = mysql_use();
        if ((row = mysql_fetch_row(res)) != NULL) {
            reply(textclient, user, "NS_NOREGISTER_REGISTERED", argv[0]);
            return;
        }
        neonserv_cmd_noregister_async1(client, textclient, user, chan, event, argv[0], duration, reason);
    } else if(argv[0][0] == '*') {
        //we've got an auth
        argv[0]++;
        printf_mysql_query("SELECT `user_user` FROM `users` WHERE `user_user` = '%s'", escape_string(argv[0]));
        res = mysql_use();
        if ((row = mysql_fetch_row(res)) != NULL) {
            neonserv_cmd_noregister_async1(client, textclient, user, chan, event, row[0], duration, reason);
        } else {
            //we need to create a new user...
            //but first lookup the auth to check if it really exists
            struct neonserv_cmd_noregister_cache *cache = malloc(sizeof(*cache));
            if (!cache) {
                perror("malloc() failed");
                return;
            }
            cache->client = client;
            cache->textclient = textclient;
            cache->user = user;
            cache->chan = chan;
            cache->event = event;
            cache->nick = strdup(argv[0]);
            cache->duration = duration;
            cache->reason = strdup(reason);
            lookup_authname(argv[0], module_id, neonserv_cmd_noregister_auth_lookup, cache);
        }
    } else {
        struct UserNode *cuser = getUserByNick(argv[0]);
        if(!cuser) {
            cuser = createTempUser(argv[0]);
			if(!cuser) {
                reply(textclient, user, "NS_USER_UNKNOWN", argv[0]);
                return;
            }
            cuser->flags |= USERFLAG_ISTMPUSER;
        }
        if(cuser->flags & USERFLAG_ISAUTHED) {
            neonserv_cmd_noregister_async1(client, textclient, user, chan, event, cuser->auth, duration, reason);
        } else {
            struct neonserv_cmd_noregister_cache *cache = malloc(sizeof(*cache));
            if (!cache) {
                perror("malloc() failed");
                return;
            }
            cache->client = client;
            cache->textclient = textclient;
            cache->user = user;
            cache->chan = chan;
            cache->event = event;
            cache->nick = strdup(user->nick);
            cache->duration = duration;
            cache->reason = strdup(reason);
            get_userauth(cuser, module_id, neonserv_cmd_noregister_nick_lookup, cache);
        }
    }
}

static AUTHLOOKUP_CALLBACK(neonserv_cmd_noregister_auth_lookup) {
    struct neonserv_cmd_noregister_cache *cache = data;
    if(!exists) {
        //AUTH_DOES_NOT_EXIST
        reply(cache->textclient, cache->user, "NS_AUTH_UNKNOWN", cache->nick);
    } else
        neonserv_cmd_noregister_async1(cache->client, cache->textclient, cache->user, cache->chan, cache->event, auth, cache->duration, cache->reason);
    free(cache->reason);
    free(cache->nick);
    free(cache);
}

static USERAUTH_CALLBACK(neonserv_cmd_noregister_nick_lookup) {
    struct neonserv_cmd_noregister_cache *cache = data;
    if(!user) {
        //USER_DOES_NOT_EXIST
        reply(cache->textclient, cache->user, "NS_USER_UNKNOWN", cache->nick);
    }
    else if(!(user->flags & USERFLAG_ISAUTHED)) {
        //USER_NOT_AUTHED
        reply(cache->textclient, cache->user, "NS_USER_NEED_AUTH", cache->nick);
    }
    else
        neonserv_cmd_noregister_async1(cache->client, cache->textclient, cache->user, cache->chan, cache->event, user->auth, cache->duration, cache->reason);
    free(cache->reason);
    free(cache->nick);
    free(cache);
}

static void neonserv_cmd_noregister_async1(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct ChanNode *chan, struct Event *event,char *auth, int duration, char *reason) {
    MYSQL_RES *res;
    MYSQL_ROW row;
    printf_mysql_query("SELECT `dnr_id` FROM `donotregister` WHERE `dnr_target` = '%s'", escape_string(auth));
    res = mysql_use();
    if((row = mysql_fetch_row(res)) != NULL) {
        printf_mysql_query("DELETE FROM `donotregister` WHERE `dnr_id` = '%s'", row[0]);
    }
    int userid;
    printf_mysql_query("SELECT `user_id` FROM `users` WHERE `user_user` = '%s'", escape_string(user->auth));
    res = mysql_use();
    if ((row = mysql_fetch_row(res)) != NULL)
        userid = atoi(row[0]);
    else
        userid = 0;
    printf_mysql_query("INSERT INTO `donotregister` (`dnr_target`, `dnr_timeout`, `dnr_user`, `dnr_reason`) VALUES ('%s', '%lu', '%d', '%s')", escape_string(auth), (duration ? (time(0)+duration) : 0), userid, escape_string(reason));
    reply(textclient, user, "NS_NOREGISTER_DONE", auth);
}

static void neonserv_cmd_noregister_list(struct ClientSocket *client, struct ClientSocket *textclient, struct UserNode *user, struct ChanNode *chan, struct Event *event, char **argv, int argc) {
    struct Table *table;
    int entrys = 0;
    MYSQL_RES *res;
    MYSQL_ROW row;
    printf_mysql_query("SELECT `dnr_target`, `dnr_timeout`, `user_user`, `dnr_reason` FROM `donotregister` LEFT JOIN `users` ON `dnr_user` = `user_id` ORDER BY `dnr_target` ASC");
    res = mysql_use();
    table = table_init(4, mysql_num_rows(res) + 1, 0);
    char *content[4];
    content[0] = get_language_string(user, "NS_DNR_TARGET");
    content[1] = get_language_string(user, "NS_DNR_USER");
    content[2] = get_language_string(user, "NS_DNR_EXPIRES");
    content[3] = get_language_string(user, "NS_DNR_REASON");
    table_add(table, content);
    int duration;
    char expires_str[MAXLEN];
    while ((row = mysql_fetch_row(res)) != NULL) {
        entrys++;
        content[0] = row[0];
        content[1] = row[2];
        duration = atoi(row[1]);
        content[2] = (duration ? timeToStr(user, (duration - time(0)), 2, expires_str) : get_language_string(user, "NS_USERS_SEEN_NEVER"));
        content[3] = row[3];
        table_add(table, content);
    }
    //send the table
    char **table_lines = table_end(table);
    int i;
    for(i = 0; i < table->entrys; i++) {
        reply(textclient, user, table_lines[i]);
    }
    if(!entrys)
        reply(textclient, user, "NS_TABLE_NONE");
    table_free(table);
}
